
#include <string>
#include <vector>
#include <cstring> // for memset()
#include <cstdlib> // for free()
#include <arpa/nameser.h> // for NS_MAXDNAME
//#include <iostream>
#include "kadmin5.hpp"
#include "context.hpp"

// awfull but necessary
extern "C"
{
#include <krb5.h>
#include <admin.h>
#include <kadm_err.h>
}


using namespace boost::python;

namespace
{
    const char*
    get_tmp_password()
    {
        // just return some data. don't want to generate it every time
        return
            "WNXErirqxBAnWNihNARIBFAR..."
            "cngnzhpugbTynqvbYHf286"
            "ZlMqrynRZFibvYhaNCNex,"
            "mOynpxWnpxbZVFUyHunZL!!!"
            ;
    }



    void
    append_realm(std::string& name, const std::string& realm)
    {
        // we need to find out wheter name contains @REALM part.
        // to do it we search for unescaped '@' symbol in it
        std::string::size_type cobako = 0;
        do {
            cobako = name.find('@', cobako + 1);
        } while ( cobako != std::string::npos && name[cobako-1] == '\\');

        if (cobako == std::string::npos)
        {
            name += '@';
            name += realm;
        }
    }
}

extern "C"
{
    // we'll use krberos internal procedure: it just does what we need
    krb5_error_code krb5int_get_fq_local_hostname (char *, size_t);
}

namespace
{
    std::string get_fq_local_hostname()
    {
        char buf[NS_MAXDNAME + 1];
        buf[0] = 0;
        krb5int_get_fq_local_hostname(buf, NS_MAXDNAME);
        buf[NS_MAXDNAME] = 0;
        return std::string(buf);
    }
}



py_kadm5::kadmin::kadmin(
          boost::python::object error_callback
        , const char* name
        , const char* realm
        )
: handle_(0), cntx_(0, error_callback), realm_(realm? realm : "")
{
    try
    {
        kadm5_config_params params;
        std::memset(&params, 0, sizeof(params));

        krb5_context cntx;
        if (krb5_error_code e =
                    kadm5_init_krb5_context(&cntx))
            throw_exception(e,"initializing krb5 library");
        cntx_.reset(cntx);

        if (realm == 0)
        {
            if (krb5_error_code e =
                    krb5_get_default_realm(context(), (char**) &realm))
                throw_exception(e, "getting default realm");
            try
            {
                realm_ = realm;
            }
            catch (...)
            {
                free((void*)realm);
                throw;
            }
            free((void*)realm);
        }

        params.realm = const_cast<char*>(realm_.c_str());
        params.mask |= KADM5_CONFIG_REALM;

        std::string fullname(name);
        append_realm(fullname, realm_);

        if (krb5_error_code e =
                    kadm5_init_with_password(
                          const_cast<char*>(name) //char *client_name,
                        , NULL                 //char *pass,
                        , NULL                 //char *service_name,
                        , &params              //kadm5_config_params *realm_params,
                        , KADM5_STRUCT_VERSION //unsigned long struct_version,
                        , KADM5_API_VERSION_2  //unsigned long api_version,
                        , NULL                 //char **db_args,
                        , &handle_             //void **server_handle
                        ))
            throw_exception(e, "initializing kadmin interface");
    }
    catch(...)
    {
        cleanup();
        throw;
    }
}


py_kadm5::kadmin::~kadmin()
{
    cleanup();
}


void
py_kadm5::kadmin::cleanup()
{
    if (handle_)
    {
        kadm5_destroy(handle_);
        handle_ = 0;

    }
}


krb5_error_code
py_kadm5::kadmin::parse_principal(std::string& name, krb5_principal* pprinc)
{
    try
    {
        append_realm(name, realm_);
        return krb5_parse_name(
                  context()
                , const_cast<char*>(name.c_str())
                , pprinc
                );
    }
    catch (...)
    {
        // All we can catch here is exceptions from std::string.
        // Current solution is not quite correct
        // (might give wrong error description to user), but safe.
    }

    return ENOMEM;
}



boost::python::list
py_kadm5::kadmin::list(const char* expr)
{
    char **names = 0;
    int count = 0;

    boost::python::list res;

    if (krb5_error_code e =
            kadm5_get_principals(
                  handle_
                , const_cast<char*>(expr)
                , &names
                , &count
                ) != 0)
        throw_exception(e, "getting principal lists");

    try
    {
        for (int i = 0; i < count; ++i)
        {
            res.append(const_cast<const char*>(names[i]));
        }
    }
    catch(...)
    {
        kadm5_free_name_list(handle_, names, count);
        throw;
    }

    kadm5_free_name_list(handle_, names, count);
    return res;
}


void
py_kadm5::kadmin::delete_princ(const char* princ_name)
{
    krb5_principal princ;
    krb5_error_code e = 0;

    std::string name(princ_name);
    if (( e = parse_principal(name, &princ) ))
        throw_exception(e, "parsing principal name");

    e = kadm5_delete_principal(handle_, princ);
    krb5_free_principal(context(), princ);
    if (e != 0)
        throw_exception(e, "deliting principal");
}

void
py_kadm5::kadmin::chpass_princ(const char* princ_name, const char* pass)
{
    krb5_principal princ;
    krb5_error_code e = 0;

    std::string name(princ_name);
    if (( e = parse_principal(name, &princ) ))
        throw_exception(e, "parsing principal name");

    e = kadm5_chpass_principal(handle_, princ, const_cast<char*>(pass));
    krb5_free_principal(context(), princ);
    if (e != 0)
        throw_exception(e, "changing principal password");
}


void
py_kadm5::kadmin::randkey_princ(const char* princ_name)
{
    krb5_principal princ;
    krb5_error_code e = 0;

    std::string name(princ_name);
    if (( e = parse_principal(name, &princ) ))
        throw_exception(e, "parsing principal name");

    e = kadm5_randkey_principal(handle_, princ, 0,0);
    krb5_free_principal(context(), princ);
    if (e != 0)
        throw_exception(e, "randomizing a key for a principal");
}

void
py_kadm5::kadmin::set_princ_attributes(const char* princ_name,
                                       krb5_flags attrs)
{
    krb5_error_code e = 0;
    kadm5_principal_ent_rec princ;
    std::memset(&princ, 0, sizeof(princ));

    std::string name(princ_name);
    if (( e = parse_principal(name, &princ.principal) ))
        throw_exception(e, "parsing principal name");

    princ.attributes = attrs;

    e = kadm5_modify_principal(
                      handle_
                    , &princ
                    , KADM5_ATTRIBUTES
                    );

    kadm5_free_principal_ent(handle_, &princ);

    if (e != 0)
        throw_exception(e, "retrieving principal attributes");
}

krb5_flags
py_kadm5::kadmin::get_princ_attributes(const char* princ_name)
{
    krb5_principal princ;
    krb5_error_code e = 0;

    std::string name(princ_name);
    if (( e = parse_principal(name, &princ) ))
        throw_exception(e, "parsing principal name");

    kadm5_principal_ent_rec princ_rec;
    std::memset(&princ_rec, 0, sizeof(princ_rec));

    e = kadm5_get_principal(
                      handle_
                    , princ
                    , &princ_rec
                    , KADM5_ATTRIBUTES
                    );

    krb5_flags res = princ_rec.attributes;
    kadm5_free_principal_ent(handle_, &princ_rec);
    krb5_free_principal(context(), princ);

    if (e != 0)
        throw_exception(e, "retrieving principal attributes");

    return res;
}


boost::python::list
py_kadm5::kadmin::get_princ_keys(const char* princ_name)
{
    krb5_principal princ;
    krb5_error_code e = 0;

    std::string name(princ_name);
    if (( e = parse_principal(name, &princ) ))
        throw_exception(e, "parsing principal name");


    boost::python::list res;

    kadm5_principal_ent_rec princ_rec;
    std::memset(&princ_rec, 0, sizeof(princ_rec));
    krb5_keyblock key;
    std::memset(&key, 0, sizeof(key));

    try
    {
        if (( e = kadm5_get_principal(
                      handle_
                    , princ
                    , &princ_rec
                    , KADM5_KEY_DATA
                    ) != 0 ))
        {
            // workaround a bug in MIT Kerberos implementation
            if (e == EPERM)
                e = KADM5_UNK_PRINC;
            throw_exception(e, "retrieving principal keys");
        }

        krb5_key_data* keys = princ_rec.key_data;
        int nkeys = princ_rec.n_key_data;

        for (int i = 0 ; i < nkeys ; ++i)
        {
            int kvno = -1;

            e = kadm5_decrypt_key(
                      handle_                   // void *server_handle
                    , &princ_rec                // kadm5_principal_ent_t entry
                    , keys[i].key_data_type[0]  // krb5_int32 ktype
                    , -1                        // krb5_int32 stype
                    , 0                         // krb5_int32 kvno
                    , &key                      // krb5_keyblock &keyblock
                    , NULL                      // krb5_keysalt &keysalt
                    , &kvno                     // int *kvnop
                    );

            if (e != 0) 
                throw_exception(e, "decripting principal key");
            res.append(
                    make_tuple(
                          kvno
                        , key.enctype
                        , str(
                              (const char*) key.contents,
                              key.length
                            )
                        //, key.length // for debugging
                        )
                    );
            krb5_free_keyblock_contents(context(), &key);
            std::memset(&key, 0, sizeof(key));
        }

    }
    catch (...)
    {
        if (e == 0) // not a kerberos error
            krb5_free_keyblock_contents(context(), &key);
        if (!PyErr_Occurred())

        {
            PyErr_SetString(PyExc_RuntimeError,
                    "unknown exception while retrieving principal keys");
        }
    }

    kadm5_free_principal_ent(handle_, &princ_rec);
    krb5_free_principal(context(), princ);

    if (PyErr_Occurred())
        throw_error_already_set();

    return res;
}


std::string
py_kadm5::kadmin::create_princ(
          const char* princ_name
        , const char* pass
        , boost::python::dict args)
{
    if (len(args) != 0)
    {
        PyErr_SetString(
                  PyExc_TypeError
                , "This version of kadmin5 does not support "
                  "additional principal creation arguments"
                );
        throw_error_already_set();
    }

    krb5_int32 mask = 0;
    kadm5_principal_ent_rec princ;
    std::memset(&princ, 0, sizeof(princ));

    std::string name(princ_name);
    if (krb5_error_code e =
            parse_principal(name, &princ.principal))
        throw_exception(e, "parsing principal name");
    mask |= KADM5_PRINCIPAL;

    krb5_error_code e =
        kadm5_create_principal(handle_, &princ, mask, const_cast<char*>(pass));
    krb5_free_principal(context(), princ.principal);
    if (e != 0)
        throw_exception(e, "creating new principal");
    return name;
}


std::string
py_kadm5::kadmin::create_service_princ(
          const char* svc
        , const char* host
        , boost::python::dict args)
{
    if (len(args) != 0)
    {
        PyErr_SetString(
                  PyExc_TypeError
                , "This version of kadmin5 does not support "
                  "additional principal creation arguments"
                );
        throw_error_already_set();
    }

    krb5_int32 mask = 0;
    kadm5_principal_ent_rec princ;
    std::memset(&princ, 0, sizeof(princ));

    std::string name(svc);

    // the host parameter is interpreted as follows:
    // if the host is non-empty string, it is desiered hostname,
    //          it's user's responsibility to enshure this is fqdn
    // if host is NULL (None), the desiered hostname is localhost's fqdn
    // if host is en empty string, no host name should be appended to principal's name
    if (!host || host[0])
    {
        name.push_back('/');

        if (!host)
            name += get_fq_local_hostname();
        else
            name += host;
    }

    if (krb5_error_code e =
            parse_principal(name, &princ.principal))
        throw_exception(e, "parsing principal name");
    mask |= KADM5_PRINCIPAL;

    // disable all tickets for the principal until we generate random key
    princ.attributes |= KRB5_KDB_DISALLOW_ALL_TIX;
    mask |= KADM5_ATTRIBUTES;

    try
    {
        if (krb5_error_code e =
            kadm5_create_principal(handle_, &princ, mask,
                   const_cast<char*>(get_tmp_password())))
            throw_exception(e, "creating new principal");

        if (krb5_error_code e =
                kadm5_randkey_principal(handle_, princ.principal, NULL, NULL))
            throw_exception(e, "generating random key for a principal");

        //now we can enable tickets
        princ.attributes &= ~KRB5_KDB_DISALLOW_ALL_TIX;
        mask = KADM5_ATTRIBUTES;
        if (krb5_error_code e =
                kadm5_modify_principal(handle_, &princ, mask))
            throw_exception(e, "enabling tickets for principal");
    }
    catch(...)
    {
        // do nothing
    }

    krb5_free_principal(context(), princ.principal);
    if (PyErr_Occurred())
        throw_error_already_set();

    return name;
}


std::string
py_kadm5::kadmin::get_realm()
{
    return realm_;
}

