
#include <cstdlib>
#include <cstring>

extern "C" {
#include <krb5.h>
}

#include "keytab.hpp"

py_kadm5::keytab::keytab(
        boost::python::object error_callback,
        const char* name)
: ctx_(error_callback), kt_(0)
{
    char keytab_name [MAX_KEYTAB_NAME_LEN + 3];

    if (!name)
    {
        // the default "FILE:" keytab type is not writable.
        // if it is really our default, replace it with WRFILE.
        name = keytab_name + 2;
        if (krb5_kt_default_name(
                      context()
                    , const_cast<char*>(name)
                    , MAX_KEYTAB_NAME_LEN
                    ) == 0)
        {
            keytab_name[MAX_KEYTAB_NAME_LEN+2] = 0;
            if (std::strncmp("FILE:",name, 5) == 0)
            {
                keytab_name[0] = 'W';
                keytab_name[1] = 'R';
                name = keytab_name;
            }
        }
        else
            name = 0;
    }

    if (name != 0)
    {
        if (krb5_error_code e =
                krb5_kt_resolve(context(), name, &kt_))
            ctx_.throw_exception(e, "while resolving keytab");
    }
    else
    {
        if (krb5_error_code e =
                krb5_kt_default(context(), &kt_))
            ctx_.throw_exception(e, "while resolving keytab");
    }
}

py_kadm5::keytab::~keytab()
{
    krb5_kt_close(context(), kt_);
}

boost::python::list
py_kadm5::keytab::list()
{
    boost::python::list res;
    krb5_error_code e = 0;

    krb5_keytab_entry ent;
    std::memset(&ent, 0, sizeof(ent));
    krb5_kt_cursor cur;
    std::memset(&cur, 0, sizeof(cur));

    if ((e = krb5_kt_start_seq_get(context(),kt_, &cur))!= 0)
        ctx_.throw_exception(e, "starting iteration over keytab");

    while ((e = krb5_kt_next_entry(context(), kt_, &ent, &cur)) == 0)
    {
        char* princname = 0;

        try
        {
            if (krb5_error_code er =
                    krb5_unparse_name(context(), ent.principal, &princname)
                    != 0 )
                ctx_.throw_exception(er, "unparsing principal name");

            res.append(
                    boost::python::make_tuple(
                          ent.vno
                        , ent.key.enctype
                        , const_cast<const char*>(princname)
                        )
                    );
            free(princname);
        }
        catch (...)
        {
            if (!PyErr_Occurred())
                PyErr_SetString(PyExc_RuntimeError,
                        "unknown exception while iterating over keytab");

            free(princname);
            krb5_free_keytab_entry_contents(context(), &ent);
            break;
        }

        krb5_free_keytab_entry_contents(context(), &ent);
        std::memset(&ent, 0, sizeof(ent)); // just in case...
    }

    krb5_kt_end_seq_get(context(),kt_, &cur);

    if (PyErr_Occurred())
        boost::python::throw_error_already_set();

    if (e != KRB5_KT_END)
        ctx_.throw_exception(e, "iterating over keytab");

    return res;
}

void
py_kadm5::keytab::add_entry(const char* princ,
        krb5_kvno kvno, krb5_enctype enctype,
        boost::python::object key
        )
{
    using namespace boost::python;

    krb5_keytab_entry ent;
    std::memset(&ent, 0, sizeof(ent));

    ent.magic = KV5M_KEYTAB_ENTRY;
    ent.vno = kvno;
    ent.key.magic = KV5M_KEYBLOCK;
    ent.key.enctype = enctype;
    ent.key.length = len(key);
    char const* key_val = extract<char const*>(key);

    if (krb5_error_code e =
            krb5_parse_name(context(), princ, &ent.principal))
        ctx_.throw_exception(e, "parsing principal name");

    // the last thing to extract
    ent.key.contents = (krb5_octet*) std::malloc(ent.key.length);
    if (! ent.key.contents)
    {
        krb5_free_principal(context(), ent.principal);
        ctx_.throw_exception(ENOMEM, "contstructing keytab entry");
    }

    std::memcpy(ent.key.contents, key_val, ent.key.length);

    krb5_error_code e = 
        krb5_kt_add_entry(context(), kt_, &ent);
    krb5_free_keytab_entry_contents(context(), &ent);
    if (e != 0)
        ctx_.throw_exception(e, "adding principal to keytab");
}


int
py_kadm5::keytab::remove_princ(const char* princ_name,
        krb5_kvno kvno, krb5_enctype enctype)
{
    int res = 0;
    krb5_principal princ;

    if (krb5_error_code e =
            krb5_parse_name(context(), princ_name, &princ))
        ctx_.throw_exception(e, "parsing principal name");

    krb5_keytab_entry ent;
    std::memset(&ent, 0, sizeof(ent));
    krb5_kt_cursor cur;
    std::memset(&cur, 0, sizeof(cur));
    krb5_error_code e = 0;

    if ((e = krb5_kt_start_seq_get(context(),kt_, &cur))!= 0)
        ctx_.throw_exception(e, "starting iteration over keytab");

    while ((e = krb5_kt_next_entry(context(), kt_, &ent, &cur)) == 0)
    {
        if (krb5_principal_compare(context(), princ, ent.principal)
                && ( kvno == 0 || ent.vno == kvno)
                && ( enctype == 0 || ent.key.enctype == enctype))
        {
            // we need to stop iteration before removing.
            // MIT kerberos kadmin also does so.
            try
            {
                if (krb5_error_code e =
                        krb5_kt_end_seq_get(context(), kt_, &cur))
                    ctx_.throw_exception(e, "interrupting iteration over keytab");

                if (krb5_error_code e =
                        krb5_kt_remove_entry(context(), kt_, &ent))
                    ctx_.throw_exception(e, "removing entry from keytab keytab");

                if (krb5_error_code e =
                        krb5_kt_start_seq_get(context(), kt_, &cur))
                    ctx_.throw_exception(e, "restarting iteration over keytab");

                ++res;
            }
            catch (...)
            {
                 krb5_free_keytab_entry_contents(context(), &ent);
                 break;
            }
        }
        krb5_free_keytab_entry_contents(context(), &ent);
        std::memset(&ent, 0, sizeof(ent)); // just in case...
    }

    krb5_kt_end_seq_get(context(),kt_, &cur);

    if (PyErr_Occurred())
        boost::python::throw_error_already_set();

    if (e != KRB5_KT_END)
        ctx_.throw_exception(e, "iterating over keytab");

    return res;
}

